#!/usr/bin/env python
"""
Test-running logic for validating tanco tests.
"""
import sys, os, errno, subprocess, traceback, json

from . import orgtest, database as db
from . import model as m
from .model import TestDescription, Config, Challenge, ResultKind, TestFailure
from .client import TancoClient

USER_HELP = """
Tanco can't find your code!

Your first step is to write a *console-mode* program
(one that does absolutely nothing!) and tell tanco
where to find it.

This is configured in your .tanco file, but you can
also use this command to test the configuration before
adding it to that file:

    tanco check [/path/to/your-program] [arguments]

The path should refer to a physical file on disk, so if
you need command line arguments, create a wrapper program.
The default path is "./my-program".

You can pass extra arguments that will be passed to the
target program.

If you need more help with setting up, try reading:
https://github.com/LearnProgramming/learntris/wiki/Getting-Set-Up
(tanco was based on the test runner for learntris)

Once tanco is able to launch your program, this message
will be replaced with instructions for implementing your
first feature.
"""


def load_config() -> Config:
    res = Config()
    res.uid = TancoClient().whoami()['id']
    res.input_path = os.environ.get("INPUT_PATH", "")
    res.skip_lines = int(os.environ.get("SKIP_LINES", "0"))
    res.test_plan = os.environ.get("TEST_PLAN")
    if os.path.exists('.tanco'):
        try:
            data = json.load(open('.tanco'))
            target = data['targets']['main']
        except json.decoder.JSONDecodeError as e:
            print("Error reading .tanco file:", e)
            sys.exit()
        except KeyError:
            print("`targets/main` found in the .tanco file.")
            sys.exit()
        if 'args' not in target:
            print("`targets/main` must specify a list of program arguments.")
            sys.exit()
        res.attempt = data.get('attempt')
        res.program_args = target['args']  # TODO: check that it's actually a list
        res.use_shell = target.get('shell', False)  # TODO: check that it's actually a bool
    return res


class NoTestPlanError(Exception):
    pass


def spawn(program_args, use_shell):
    try:
        return subprocess.Popen(program_args,
                            shell=use_shell,
                            universal_newlines=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE)
    except OSError as e:
        if e.errno == errno.ENOENT:
            print("Couldn't find program:", program_args[0])
            print("Make sure you have the right path.")
        else:
            print(f"Couldn't run program '{program_args[0]}':")
            print('  ', e)
        sys.exit()


def send_cmds(cfg: Config, program, ilines):
    if cfg.input_path:
        cmds = open(cfg.input_path, "w")
        for cmd in ilines:
            cmds.write(cmd + "\n")
        cmds.close()
    else:
        for cmd in ilines:
            program.stdin.write(cmd + "\n")
            program.stdin.flush()
        program.stdin.close()


def run_test(cfg: Config, program: subprocess.Popen, test: TestDescription):
    # send all the input lines (to stdin or a file):
    send_cmds(cfg, program, test.ilines)
    # listen for the response:
    (actual, _errs) = program.communicate(timeout=5)
    # TODO: handle errors in the `errs` string
    actual = clean_output(cfg, actual)
    local_check_output(cfg, actual, test)


def clean_output(cfg: Config, actual: [str]) -> [str]:
    actual = [line.strip() for line in actual.splitlines()]
    actual = actual[cfg.skip_lines:]
    # strip trailing blank lines
    while actual and actual[-1] == "":
        actual.pop()
    return actual


def save_new_rule(attempt: str, test: str, rule: m.ValidationRule):
    db.save_progress(attempt, test, True)
    db.save_rule(attempt, test, rule.to_data())


def local_check_output(cfg: m.Config, actual: [str], test: TestDescription):
    local_res = test.check_output(actual)
    match local_res.kind:
        case ResultKind.Pass: pass
        case ResultKind.Fail:
            # a regression! (test failed and have the rule,
            # so we have to tell the server)
            fail(cfg, local_res.error.error_lines(), test.name, local_res)
        case ResultKind.AskServer:
            client = TancoClient()
            remote_res = client.check_output(cfg.attempt, test.name, actual)
            match remote_res.kind:
                case ResultKind.Pass:
                    save_new_rule(cfg.attempt, test.name, remote_res.rule)
                case ResultKind.Fail:
                    raise remote_res.error
                case ResultKind.AskServer:
                    raise RecursionError("Server validation loop")


def get_challenge(cfg: Config) -> Challenge:
    if cfg.test_plan:
        return orgtest.read_challenge(cfg.test_plan)
    elif cfg.attempt:
        return db.challenge_from_attempt(cfg.attempt)
    else:
        raise NoTestPlanError()


def run_tests(cfg: Config):
    num_passed = 0
    challenge = get_challenge(cfg)
    tests = challenge.tests
    try:
        for i, test in enumerate(tests):
            program = spawn(cfg.program_args, cfg.use_shell)
            run_test(cfg, program, test)
            # either it passed or threw exception
            print('.', end='')
            num_passed += 1
        else:
            print()
            print("All %d tests passed." % num_passed)
            print()
            print("This may be a good time to commit your changes,")
            print("or spend some time improving your code.")
            print()
            print("When you're ready, run `tanco next` to start work on the next feature.")
            print()
            TancoClient().send_pass(cfg.attempt)
    except (subprocess.TimeoutExpired, TestFailure) as e:
        print()
        print("%d of %d tests passed." % (num_passed, len(tests)))
        if isinstance(e, subprocess.TimeoutExpired):
            print("Test [%s] timed out." % test.name)
        else:
            print("Test [%s] failed." % test.name)
            print()
            print('#', test.name, test.head)
            print(test.body)
            print(" --- input given ------")
            for line in test.ilines:
                print(line)
            print()
            fail(cfg, e.error_lines(), test)


def find_target(cfg: Config, argv: [str]) -> Config:
    """returns the config, possibly overridden by command line args"""
    if len(argv) > 1:
        cfg.program_args = argv[1:]
        if "--shell" in cfg.program_args:
            cfg.program_args.remove("--shell")
            cfg.use_shell = True
    if cfg.use_shell or os.path.exists(cfg.program_args[0]):
        return cfg
    elif cfg.program_args[0] == cfg.default_target():
        print(USER_HELP)
    raise FileNotFoundError(cfg.program_args[0])


def get_custom_cfg(argv: [str]):
    """return a config object, possibly overridden by command line args"""
    cfg = load_config()
    try:
        cfg = find_target(cfg, argv)
    except FileNotFoundError as e:
        print('File not found:', e)
        sys.exit()
    return cfg


def check(argv: [str]):
    cfg = get_custom_cfg(argv)
    program = spawn(cfg.program_args, cfg.use_shell)
    test = TestDescription()
    test.name = 'check'
    test.head = 'tanco check'
    test.body = '\n'.join([
        "Tanco needs to be able to run your program.",
        "Please make sure that your program is marked as executable,",
        "and that by default, it produces no output and returns exit code 0"
    ])
    try:
        run_test(cfg, program, test)
    except SystemExit:
        raise
    except:
        handle_unexpected_error(cfg)
    else:
        print(f"Tanco ran {' '.join(cfg.program_args)} successfully.")
        print("Run `tanco next` to start the first test.")


def fail(cfg: Config, msg: [str], tn: str = None, tr: m.TestResult = None):
    for line in msg:
        print(line)
    if tn and tr:
        assert tr.kind == ResultKind.Fail, "Expected a failed test."
        c = TancoClient()
        c.send_fail(cfg.attempt, tn, tr)
    sys.exit()


def handle_unexpected_error(cfg: Config):
    try:
        fail(cfg, ['-'*50,
                   traceback.format_exc(),
                   '-'*50,
                   "Oh no! Tanco encountered an unexpected problem while",
                   "attempting to run your program. Please report the above",
                   "traceback in the issue tracker, so that we can help you",
                   "with the problem and provide a better error message in",
                   "the future.",
                   "",
                   "  https://github.com/tangentcode/tanco/issues"])
    except Exception as e:
        traceback.print_exception(type(e), e, None)
        sys.exit()


def main(argv: [str]):
    cfg = get_custom_cfg(argv)
    cmdline = cfg.program_args
    cmd = cmdline[0]
    try:
        run_tests(cfg)
    except SystemExit:
        pass
    except NoTestPlanError:
        fail(cfg, ['No challenge selected.'
                   'Use `tanco init` or set TEST_PLAN environment variable.'])
    except EnvironmentError as e:
        if e.errno in [errno.EPERM, errno.EACCES]:
            fail(cfg, [str(e),
                       "Couldn't run %r due to a permission error." % cmd,
                       "Make sure your program is marked as an executable."])
        elif e.errno == errno.EPIPE:
            fail(cfg, [str(e),
                       "%r quit before reading any input." % cmd,
                       "Make sure you are reading commands from standard input,",
                       "not trying to take arguments from the command line."])
        else:
            handle_unexpected_error(cfg)
    except:
        handle_unexpected_error(cfg)


if __name__ == '__main__':
    main(sys.argv)



import sqlite3
from . import model as m
import os

SDB_PATH = os.environ.get('TANCO_SDB_PATH')
if not SDB_PATH:
    SDB_PATH = os.path.expanduser('~/.tanco.sdb')


def ensure_sdb():
    if not os.path.exists(SDB_PATH):
        print("Creating database at", SDB_PATH)
        import tanco
        sql = open(os.path.join(*tanco.__path__ + ['sql', 'init.sql'])).read()
        dbc = begin()
        dbc.executescript(sql)
        dbc.commit()


def query(sql, *a, **kw) -> [{}]:
    """fetch a relation from the database"""
    dbc = sqlite3.connect(SDB_PATH)
    cur = dbc.execute(sql, *a, **kw)
    cols = [x[0] for x in cur.description]
    return [{k: v for k, v in zip(cols, vals)}
            for vals in cur.fetchall()]


def commit(sql, *a, **kw) -> int | None:
    """commit a transaction to the database"""
    dbc = begin()
    cur = dbc.execute(sql, *a, **kw)
    dbc.commit()
    return cur.lastrowid


def begin() -> sqlite3.Connection:
    """return a connection so you can begin a transaction"""
    tx = sqlite3.connect(SDB_PATH)
    tx.execute('PRAGMA foreign_keys = ON')
    return tx


def chomp(lines: [str]) -> [str]:
    """remove trailing blank line"""
    if not lines: return []
    return lines[:-1] if lines[-1] == '' else lines


def fetch_challenge(chid: int):
    """fetch a challenge from the database"""
    rows = query('select * from challenges where id=?', [chid])
    if not rows:
        raise LookupError(f'Challenge "{chid}" not found in the database.')
    res = m.Challenge(**rows[0])
    for row in query('select * from tests where chid=?', [chid]):
        res.tests.append(test_from_row(row))
    return res


def test_from_row(row) -> m.TestDescription:
    t = row
    t['ilines'] = chomp(t['ilines'].split('\n'))
    if t['olines'] is not None:
        t['olines'] = chomp(t['olines'].split('\n'))
    return m.TestDescription(**t)


def challenge_from_attempt(aid: str):
    """fetch a challenge from the database"""
    rows = query('select chid from attempts where code=(:code)',
                 {'code': aid})
    if not rows:
        raise LookupError(f'Attempt "{aid}" not found in the database.')
    return fetch_challenge(rows[0]['chid'])


def get_server_id(url):
    """get the server id for a given url"""
    rows = query('select id from servers where url=?', [url])
    if not rows:
        raise LookupError(f'Server "{url}" not found in the database.')
    return rows[0]['id']


def get_next_tests(aid: str, uid: int):
    """get the next group of tests for a given attempt"""
    return query("""
        select t.* from (
            select t.chid, t.grp
            from (attempts a left join tests t on a.chid=t.chid)
                left join progress p on t.id=p.tid
            where a.code=(:code) and a.uid = (:uid)
            group by t.grp having count(p.id)=0
            order by t.grp limit 1) as g
        left join tests t on g.chid=t.chid and g.grp=t.grp
        """, {'code': aid, 'uid': uid})


def save_progress(attempt: str, test: str, _passed: bool):
    """save progress when a test passes"""
    commit("""
      insert into progress (aid, tid)
        select a.id as aid, t.id as tid
        from attempts a, tests t
        where a.chid = t.chid
          and a.code = ? and t.name = ?
        """, [attempt, test])


def save_rule(attempt: str, test: str, rule: dict):
    """save a rule for a test"""
    # TODO: save the rule to the database as json
    # json.dumps(rule)

    # for now, we still have this 'olines' thing
    assert rule['kind'] == 'lines', \
        f"don't know how to save {rule['kind']!r} rules"
    commit("""
        update tests set olines = ?
        where name = ?
          and chid = (
            select a.chid from attempts a
            where a.code = ?)
        """, ['\n'.join(rule['data']), test, attempt])


def get_attempt_test(uid, code, test_name):
    rows = query(
        """
        select t.name, t.head, t.body, t.ilines, t.olines
        from attempts a left join tests t on a.chid=t.chid
        where a.code=? and a.uid=? and t.name=?
        """, [code, uid, test_name])
    # the actual output from the test run is the request body (json)
    if not rows:
        raise LookupError(f'attempt: {code}, test: {test_name}')
    return test_from_row(rows[0])


def set_attempt_state(uid, code, transition: m.Transition, failing_test: str = '') -> (m.AttemptState, str):
    """set the state of an attempt according to transition table"""
    try:
        old = query("""
            select a.id as aid, a.state, t.name as focus
            from attempts a left join tests t on a.focus = t.id
            where a.code=? and a.uid=?""", [code, uid])[0]
    except IndexError:
        raise LookupError(f'attempt: {code}')

    # o: one-letter code for old state:
    # 'sbfcd' start build fix change done (the possible states)
    o = old['state'][0].lower()

    new_focus = None

    # t: one-letter code for transition:
    # 'XPNO' X:tanco-next P=test pass, N=new fail, O=old fail
    match transition:
        case m.Transition.Pass: t = 'P'
        case m.Transition.Next: t = 'N'  # !! what about `tanco next` but no more tests?
        case m.Transition.Fail:
            assert failing_test, "failing test required for Fail transition"
            row = query("""
                select t.id, count(p.id) > 0 as is_regression
                from attempts a, tests t left join progress p on t.id = p.tid
                where a.id = (:aid) and a.chid = t.chid
                  and a.id = p.aid and t.name = (:test)
                group by t.id""", {'aid': old['aid'], 'test': failing_test})[0]
            new_focus, is_regression = row['id'], row['is_regression']
            t = 'O' if is_regression else 'F'
        case _: raise ValueError(f"unknown transition: {transition}")

    # state machine transition table:
    sm = {'s': {'X': 'b'},
          'b': {'P': 'c', 'N': 'b', 'O': 'f'},
          'f': {'O': 'f', 'N': 'b', 'P': 'c'},
          'c': {'X': '?', 'O': 'f', 'P': 'c'},
          'd': {'X': 'd', 'O': 'f', 'P': 'd'}}

    # 's.X:b'  # others can't happen (no test until build state)
    # 's.P:s'  # you pass the 0 tests at the start
    # 'b.X->ERR'
    # 'b.P:c'
    # 'b.N:b'
    # 'b.O:f'
    # 'f.O:f'
    # 'f.N:b'
    # 'f.P:c'
    # 'c.P:c'
    # 'c.N:f'  # really this can't happen because no 'new' test anymore
    # 'c.O:f'  # all tests are old tests
    # 'c.X:(b|d)'
    # 'd.X:d'  # nothing more to do
    # 'd.P:d'  # you ran the tests just to see them pass
    # 'd.O:f'  # you put yourself back in 'change' mode without telling us

    # n: one-letter code for new state (same as codes for o)
    n = sm[o].get(t)
    if not n:
        raise ValueError(f"invalid transition: {o}.{t}")
    elif t == '?':  # c.X ('tanco next' from 'change' state)
        next_tests = get_next_tests(code, uid)
        if next_tests:
            n = 'b'
            new_focus = next_tests[0]['id']
        else:
            n = 'd'

    match n:
        case 's': new_state = m.AttemptState.Start
        case 'b': new_state = m.AttemptState.Build
        case 'f': new_state = m.AttemptState.Fix
        case 'c': new_state = m.AttemptState.Change
        case 'd': new_state = m.AttemptState.Done
        case _: raise ValueError(f"unknown state: {n}")

    commit("""
        update attempts set state=(:new_state), focus=(:new_focus)
        where id=(:aid)
        """, {'new_state': new_state.name.lower(),
              'new_focus': new_focus, 'aid': old['aid']})

    return new_state, failing_test


def uid_from_tokendata(sid, authid, username):
    rows = query("select id from users where sid=? and authid=?", [sid, authid])
    if rows:
        uid = rows[0]['id']
    else:
        uid = commit("""
          insert into users (sid, authid, username)
          values (?, ?, ?)
          """, [sid, authid, username])
    return uid
